
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from numereval.scores import score_signals


def calculate_max_drawdown(validation_correlations: pd.Series):

    rolling_max = (
        (validation_correlations + 1).cumprod().rolling(window=100, min_periods=1).max()
    )
    daily_value = (validation_correlations + 1).cumprod()
    max_drawdown = -((rolling_max - daily_value)).max()

    #TODO: Add Max drawdown weeks
    return max_drawdown

def run_analytics(era_scores, roll_mean=20, plot=False):

    '''Calculates some stats and plot cumulative scores.
        Taken from Jason Rosenfeld's notebook.
    '''

    metrics = {}

    metrics["weeks"] = len(era_scores)
    metrics["Mean correlation"] = era_scores.mean()
    metrics["Median correlation"] = era_scores.median()
    metrics["Std. Dev."] = era_scores.std()
    metrics["Mean Pseudo-Sharpe"] = era_scores.mean()/era_scores.std()
    metrics["Median Pseudo-Sharpe"] = era_scores.median()/era_scores.std()
    metrics["Hit Rate (% positive eras)"] = era_scores.apply(lambda x: np.sign(x)).value_counts()[1]/len(era_scores) * 100
    #metrics["Max Drawdown"] = calculate_max_drawdown(era_scores)

    if plot:

        era_scores.rolling(roll_mean).mean().plot(
            kind="line", title="Rolling Per Era Correlation Mean", figsize=(15, 4)
        )
        plt.axhline(y=0.0, color="black", linestyle="-", linewidth=1)
        plt.axhline(y=era_scores.mean(), color="g", linewidth=1, linestyle="--", label='Mean corr')
        plt.show()

        era_scores.cumsum().plot(title="Cumulative Sum of Era Scores", figsize=(15, 4))
        plt.axhline(y=0.0, color="r", linestyle="--")
        plt.show()
    
    return pd.Series(metrics).round(4)


import numpy as np
import pandas as pd
import scipy

from numereval.scores import *


def neutralize(
    df, columns, extra_neutralizers=None, proportion=1.0, normalize=True, era_col="era"
):
    # need to do this for lint to be happy bc [] is a "dangerous argument"
    if extra_neutralizers is None:
        extra_neutralizers = []
    unique_eras = df[era_col].unique()
    computed = []
    for u in unique_eras:
        print(u, end="\r")
        df_era = df[df[era_col] == u]
        scores = df_era[columns].values
        if normalize:
            scores2 = []
            for x in scores.T:
                x = (pd.Series(x).rank(method="first").values - 0.5) / len(x)
                scores2.append(x)
            scores = np.array(scores2).T
            extra = df_era[extra_neutralizers].values
            exposures = np.concatenate([extra], axis=1)
        else:
            exposures = df_era[extra_neutralizers].values

        scores -= proportion * exposures.dot(
            np.linalg.pinv(exposures.astype(np.float32)).dot(scores.astype(np.float32))
        )

        scores /= scores.std()

        computed.append(scores)

    return pd.DataFrame(np.concatenate(computed), columns=columns, index=df.index)


# to neutralize any series by any other series
def neutralize_series(series, by, proportion=1.0):
    scores = series.values.reshape(-1, 1)
    exposures = by.values.reshape(-1, 1)

    # this line makes series neutral to a constant column so that it's centered and for sure gets corr 0 with exposures
    exposures = np.hstack(
        (exposures, np.array([np.mean(series)] * len(exposures)).reshape(-1, 1))
    )

    correction = proportion * (
        exposures.dot(np.linalg.lstsq(exposures, scores, rcond=None)[0])
    )
    corrected_scores = scores - correction
    neutralized = pd.Series(corrected_scores.ravel(), index=series.index)
    return neutralized


def unif(df):
    x = (df.rank(method="first") - 0.5) / len(df)
    return pd.Series(x, index=df.index)


def get_feature_neutral_mean(df):
    feature_cols = [c for c in df.columns if c.startswith("feature")]
    df.loc[:, "neutral_sub"] = neutralize(df, [PREDICTION_NAME], feature_cols)[
        PREDICTION_NAME
    ]
    scores = (
        df.groupby("era")
        .apply(lambda x: correlation(x["neutral_sub"], x[TARGET_NAME]))
        .mean()
    )
    return np.mean(scores)



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from numereval.scores import *
from numereval.neutralize import *
from typing import Union


def calculate_val_corrs(validation_data):

    validation_correlations = validation_data.groupby("era").apply(score)
    return validation_correlations


def calculate_sharpe(validation):

    validation_sharpe = validation_correlations.mean() / validation_correlations.std(
        ddof=0
    )
    return validation_sharpe


def plot_correlation(validation_correlations: pd.Series, diagnostics=False):

    if diagnostics:

        colors = []
        era_num = validation_correlations.index.str.slice(start=3).astype(int) 
        for e in era_num:
            if e in range(121, 132+1):
                colors.append("#007691")
            elif e in range(197, 206+1):
                colors.append("#50bf84")
            elif e in range(207, 212+1):
                colors.append("#3ac4e1")

        r_patch = mpatches.Patch(color="#007691", label="Val1")
        g_patch = mpatches.Patch(color="#50bf84", label="Val2")
        b_patch = mpatches.Patch(color="#3ac4e1", label="Val3")

        fig, ax = plt.subplots()
        validation_correlations.plot(
            kind="bar", color=colors, legend=True, title="Validation correlations"
        )
        ax.legend(handles=[r_patch, g_patch, b_patch])
        plt.axhline(y=validation_correlations.mean(), 
                linewidth=1,
                color='b', linestyle='-')
        plt.show()
    else:
        fig, ax = plt.subplots()
        validation_correlations.plot(kind="bar", title="Correlations")
        plt.axhline(y=validation_correlations.mean(), 
                linewidth=1,
                color='r', linestyle='-')
        plt.show()


def calculate_max_drawdown(validation_correlations: pd.Series):

    rolling_max = (
        (validation_correlations + 1).cumprod().rolling(window=100, min_periods=1).max()
    )
    daily_value = (validation_correlations + 1).cumprod()
    max_drawdown = -(rolling_max - daily_value).max()

    return max_drawdown


def calculate_feature_exposure(validation_data):

    feature_names = [f for f in validation_data.columns if f.startswith("feature")]
    feature_exposures = validation_data[feature_names].apply(
        lambda d: correlation(validation_data[PREDICTION_NAME], d), axis=0
    )

    feature_exposure = np.sqrt(np.mean(np.square(feature_exposures)))

    max_per_era = validation_data.groupby("era").apply(
        lambda d: d[feature_names].corrwith(d[PREDICTION_NAME]).abs().max()
    )
    max_feature_exposure = max_per_era.mean()

    return max_feature_exposure, feature_exposure


def evaluate(validation_data: pd.DataFrame, plot=False, feature_exposure=False,
    eras= None
):

    feature_names = [f for f in validation_data.columns if f.startswith("feature")]

    if eras is not None:
        validation_data = validation_data[validation_data.era.isin(eras)]

    metrics = {}

    validation_correlations = calculate_val_corrs(validation_data)
    if plot:
        plot_correlation(validation_correlations)

    metrics["mean"] = validation_correlations.mean()
    metrics["std"] = validation_correlations.std(ddof=0)
    validation_sharpe = metrics["mean"] / metrics["std"]
    metrics["sharpe"] = validation_sharpe
    metrics["max_drawdown"] = calculate_max_drawdown(validation_correlations)

    if feature_exposure:
        (
            metrics["max_feature_exp"],
            metrics["feature_exposure"],
        ) = calculate_feature_exposure(validation_data)

    return pd.DataFrame.from_dict(metrics, orient="index",  columns=["metrics"]).round(4)


def diagnostics(
    validation_data: pd.DataFrame, plot=False, example_preds_loc: str = None,
    eras= None
):

    if example_preds_loc is None:
        example_preds_loc = "example_predictions.csv"

    feature_names = [f for f in validation_data.columns if f.startswith("feature")]

    if eras is not None:
        validation_data = validation_data[validation_data.era.isin(eras)]

    metrics = {}

    validation_correlations = calculate_val_corrs(validation_data)
    if plot:
        plot_correlation(validation_correlations, diagnostics=True)

    metrics["mean"] = validation_correlations.mean()
    metrics["std_dev"] = validation_correlations.std(ddof=0)
    validation_sharpe = metrics["mean"] / metrics["std_dev"]
    metrics["sharpe"] = validation_sharpe

    metrics["feat_neutral_mean"] = get_feature_neutral_mean(validation_data)

    metrics["max_drawdown"] = calculate_max_drawdown(validation_correlations)

    (
        metrics["max_feature_exp"],
        metrics["feature_exposure_mean"],
    ) = calculate_feature_exposure(validation_data)

    example_preds = pd.read_csv(example_preds_loc).set_index("id")["prediction"]
    validation_example_preds = example_preds.loc[validation_data.index]
    validation_data["ExamplePreds"] = validation_example_preds

    scores = validation_data.groupby("era").apply(
        lambda x: correlation(unif(x[PREDICTION_NAME]), 
                            x["ExamplePreds"])
        )

    mmc_scores = []
    corr_scores = []
    for _, x in validation_data.groupby("era"):
        series = neutralize_series(
            pd.Series(unif(x[PREDICTION_NAME])), pd.Series(unif(x["ExamplePreds"]))
        )
        mmc_scores.append(np.cov(series, x[TARGET_NAME])[0, 1] / (0.29 ** 2))
        corr_scores.append(correlation(unif(x[PREDICTION_NAME]), x[TARGET_NAME]))

    val_mmc_mean = np.mean(mmc_scores)
    val_mmc_std = np.std(mmc_scores)
    val_mmc_sharpe = val_mmc_mean / val_mmc_std
    corr_plus_mmcs = [c + m for c, m in zip(corr_scores, mmc_scores)]
    corr_plus_mmc_sharpe = np.mean(corr_plus_mmcs) / np.std(corr_plus_mmcs)
    corr_plus_mmc_mean = np.mean(corr_plus_mmcs)
    corr_plus_mmc_sharpe_diff = corr_plus_mmc_sharpe - validation_sharpe

    metrics["mmc_mean"] = val_mmc_mean
    metrics["corr_plus_mmc_sharpe"] = corr_plus_mmc_sharpe
    #metrics["corr_plus_mmc_diff"] = corr_plus_mmc_sharpe_diff
    metrics["corr_example_preds"] = scores.mean()
    return pd.DataFrame.from_dict(metrics, orient="index", columns=["metrics"]).round(4)



import numpy as np
import pandas as pd

TOURNAMENT_NAME = "nomi"
TARGET_NAME = f"target"
PREDICTION_NAME = f"prediction"
PREDICTION_NAME_S = "signal"


def correlation(predictions, targets):
    ranked_preds = predictions.rank(pct=True, method="first")
    return np.corrcoef(ranked_preds, targets)[0, 1]


# convenience method for scoring
def score(df):
    return correlation(df[PREDICTION_NAME], df[TARGET_NAME])


def score_signals(df):
    return correlation(df[PREDICTION_NAME_S], df[TARGET_NAME])


# Payout is just the score cliped at +/-25%
def payout(scores):
    return scores.clip(lower=-0.25, upper=0.25)
